/*
SCARF Association Rule Mining Program
Copyright (c) 2012-13 Balazs Szalkai

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "logicalexpr.hpp"

#include <algorithm>
#include "mine.hpp"
#include <string>
#include "table.hpp"
#include "utils.hpp"
#include <vector>

void ConjunctiveNormalForm::createFromPattern(const std::vector<size_t>& orClauseSizes)
{
    nodes.clear();
    for (size_t n : orClauseSizes)
    {
        for (size_t i = 0; i < n; ++i)
            nodes.push_back(Node(i == n-1 ? Operator::And : Operator::Or));
    }
}

void ConjunctiveNormalForm::parsePattern(const std::string &pattern, bool sort)
{
    std::vector<std::string> tokens = split(pattern, isWhitespace, EmptyEntryAction::Remove);

	std::vector<size_t> orClauseSizes(tokens.size());
	std::transform(tokens.begin(), tokens.end(), orClauseSizes.begin(),
        [](const std::string& s){return static_cast<size_t>(parseInt(s));});
    if (sort) std::sort(orClauseSizes.begin(), orClauseSizes.end());

	createFromPattern(orClauseSizes);
}

std::vector<Range<size_t>> ConjunctiveNormalForm::getPattern() const
{
    std::vector<Range<size_t>> result;
    size_t start = 0;
    for (size_t i = 0, n = nodes.size(); i < n; ++i)
    {
        const Node& node = at(nodes, i);
        if (node.op == Operator::And || i == n-1)
        {
            result.emplace_back(start, i+1);
            start = i+1;
        }
    }
    return result;
}

std::int64_t ConjunctiveNormalForm::getHash() const
{
    std::vector<size_t> activeOrClause;
    std::vector<int64_t> orClauseHashes;

    for (size_t i = 0, n = nodes.size(); i < n; ++i)
    {
        const Node& node = at(nodes, i);
        if (!node.column.isNull())
            activeOrClause.push_back((size_t)node.column);

        if ((i == n-1 || node.op == Operator::And) && activeOrClause.size() > 0)
        {
            std::sort(activeOrClause.begin(), activeOrClause.end());
            std::int64_t h = HASH_NULL;
            for (size_t column : activeOrClause)
                h = (h ^ column) * HASH_PRIME;
            activeOrClause.clear();
            orClauseHashes.push_back(h);
        }
    }

    std::sort(orClauseHashes.begin(), orClauseHashes.end());
    std::int64_t result = HASH_NULL;
    for (int64_t h : orClauseHashes)
        result = (result ^ h) * HASH_PRIME;
    return result;
}

std::string ConjunctiveNormalForm::toString(const Mine& mine) const
{
    std::string result = "(";

    Nullable<Operator> op;
    for (const Node& node : nodes)
    {
        if (!node.column.isNull())
        {
            if (!op.isNull())
                result += (op == Operator::And ? ") & (" : " or ");
            result += at(mine.columns(), (size_t)node.column).name;
            op = node.op;
        }
        else if (op == Operator::Or && node.op == Operator::And) op = Operator::And;
    }

    return result + ")";
}

void ConjunctiveNormalForm::toXml(XmlWriter& xml, const Mine &mine) const
{
    const std::vector<Range<size_t>> orClauses = getPattern();
    xml.start("and");
    for (const Range<size_t>& range : orClauses)
    {
        // size of OR clause?
        size_t orClauseSize = 0;
        for (size_t i : range)
            if (!at(nodes, i).column.isNull())
        {
            ++orClauseSize;
            if (orClauseSize >= 2)
                break;
        }
        if (orClauseSize == 0) continue;

        if (orClauseSize >= 2)
            xml.start("or");
        for (size_t i : range)
        {
            const Node& node = at(nodes, i);
            if (node.column.isNull())
                continue;
            xml.start("variable");
            xml.text(at(mine.columns(), (size_t)node.column).name);
            xml.end();
        }
        if (orClauseSize >= 2)
            xml.end();
    }
    xml.end();
}
