/*
SCARF Association Rule Mining Program
Copyright (c) 2012-13 Balazs Szalkai

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

// Define this if you want multithreaded operation.
#define CONCURRENT

// The URL of the stylesheet for the XML output.
static const char* XML_OUTPUT_STYLESHEET = "http://pitgroup.org/scarf.xsl";
static const char* SCARF_PROGRAM_VERSION = "1.0";

#include <algorithm>
#include "concurrent_stream.hpp"
#include <cstdio>
#include <cstdlib>
#include <fstream>
#include <iostream>
#include "mine.hpp"
#include "miner.hpp"
#include "params.hpp"
#include <string>
#include "table.hpp"
#include "utils.hpp"
#include <vector>

#ifdef CONCURRENT
    #include "jobpool.hpp"
    #include "tinythread.h"
#endif

ConcurrentStream logger(std::cerr);

template <class Out>
void printUsage(const Out& out, const std::string& programName)
{
    out<<"Usage: "<<programName<<" <parameter file>"<<std::endl;
	out<<"By default, the number of threads equals to the number of CPUs."<<std::endl;
}

template <class Out>
void printAbout(const Out& out)
{
	out<<"SCARF Simple Combinatorial Association Rule Finder, v"<<SCARF_PROGRAM_VERSION<<std::endl;
	out<<"Copyright (c) 2012-13 Balazs Szalkai"<<std::endl;
	out<<"This program comes with ABSOLUTELY NO WARRANTY."<<std::endl;
	out<<"This program is released under version 3 of the GNU General Public License."<<std::endl;
    out<<"You should have received a copy of the GNU General Public License along with this program."
        " If not, see <http://www.gnu.org/licenses/>."<<std::endl;

#ifndef NDEBUG
	out<<"*** Debug mode ***"<<std::endl;
#endif
	out<<std::endl;
}

int run(const std::string& programName, const std::vector<std::string>& args)
{
	printAbout(logger);

    const size_t expectedArgCount = 1;
    if (args.size() != expectedArgCount)
	{
        logger<<"Error: wrong number of arguments: got "<<args.size()<<", expected "<<expectedArgCount<<std::endl;
	    printUsage(logger, programName);
		return 1;
	}

    // param file
	const std::string &parameterFile = at(args, 0);
	Params params(parameterFile, logger);

#ifdef CONCURRENT
	if (params.nThreads > 0)
    {
        logger<<"Number of threads given as "<<params.nThreads<<std::endl;
    }
    else
    {
        logger<<"Invalid or missing nThreads (number of threads) parameter"<<std::endl;
        params.nThreads = tthread::thread::hardware_concurrency();
        if (params.nThreads > 0)
        {
            logger<<"Launching "<<params.nThreads<<" threads (number of CPUs)"<<std::endl;
        }
        else
        {
            params.nThreads = 4; // assume at most four CPUs
            logger<<"Could not determine CPU count, launching "<<params.nThreads<<" threads"<<std::endl;
        }
    }
#else
    logger<<"Single-threaded execution"<<std::endl;
#endif

	logger<<"loading table "<<params.tableFile<<"..."<<std::endl;
	Table table;
	try
	{
        table.load(params.tableFile);
	}
	catch (const std::exception &)
	{
		logger<<"Error: table file is in an incorrect format."<<std::endl;
		return 2;
	}
	logger<<table.rows.size()<<" rows"<<std::endl;
	if (table.rows.size() == 0)
    {
        logger<<"Table file could not be opened or contained an empty table."<<std::endl;
        return 2;
    }
    for (const TableColumn& column : table.columns)
    {
        if (!Table::validColumnName(column.name))
        {
            logger<<"Error: invalid column name: "<<column.name<<std::endl;
            logger<<"Column names must match the regex ^[a-zA-Z0-9_\\.\\[\\]]+$"<<std::endl;
            return 2;
        }
        if (column.hadLongCells)
            logger<<"Warning: column '"<<column.name<<"' has cells longer than one character. "
                "Only the first character was considered."<<std::endl;
    }

	logger<<std::endl<<"SCARF is mining..."<<std::endl;
	Mine mine(table, params);

	BestRules bestRules(params.maxRuleCount);
	{
        RuleMiner miner(mine, params, logger);

        std::vector<size_t> rhsColumnIndexes = miner.get_rhsColumnIndexes();
        logger<<"Number of RHS columns: "<<rhsColumnIndexes.size()<<std::endl;
        // sort RHS columns by support _descending_ --> this will improve scheduling efficiency
        std::sort(rhsColumnIndexes.begin(), rhsColumnIndexes.end(),
            [&](const size_t a, const size_t b) { return at(mine.columns(), a).support > at(mine.columns(), b).support; });

#ifdef CONCURRENT
        JobPool<size_t, BestRules, RuleMiner> jobPool(miner, params.nThreads);
        for (const size_t rhsIndex : rhsColumnIndexes)
            jobPool.push(rhsIndex);

        BestRules rulesForRhs;
        while (jobPool.waitAndPop(rulesForRhs))
            bestRules.merge(rulesForRhs);
#else
        for (const size_t rhsIndex : rhsColumnIndexes)
            bestRules.merge(miner(rhsIndex));
#endif
	}

	// output
	logger<<std::endl;

    if (params.outputFile.size() > 0)
        std::freopen(params.outputFile.c_str(), "w", stdout);
    for (const Rule& r : bestRules.items())
        std::cout<<r.toString(mine)<<std::endl;

    if (params.xmlOutputFile.size() > 0)
    {
        std::ofstream f(params.xmlOutputFile);
        XmlWriter xml(f);
        f<<"<?xml-stylesheet type=\"text/xsl\" href=\""<<XML_OUTPUT_STYLESHEET<<"\"?>"<<std::endl;
        xml.start("rules");
        xml.text("\n");
        for (const Rule& r : bestRules.items())
        {
            r.toXml(xml, mine);
            xml.text("\n");
        }
        xml.end();
        xml.text("\n");
    }

    size_t n = bestRules.items().size();
    switch (n)
    {
        case 0: logger<<"No rules were found"<<std::endl; break;
        case 1: logger<<"1 rule was written"<<std::endl; break;
        default: logger<<n<<" rules were written"<<std::endl; break;
    }

    return 0;
}

int main(int argc, char** argv)
{
    DefaultPopCount::init();
    return run(*argv, std::vector<std::string>(argv + 1, argv + argc));
}
