/*
SCARF Association Rule Mining Program
Copyright (c) 2012-13 Balazs Szalkai

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "mine.hpp"

#include <algorithm>
#include <cstdlib>
#include <fstream>
#include <iostream>
#include "table.hpp"
#include "utils.hpp"

MineColumn::MineColumn(const MineColumn& other, const std::set<size_t>& keepRows)
    : MineColumn(other, DontCopyData())
{
    block_t bufferTrue = Zeros<block_t>(), bufferNotNA = Zeros<block_t>();
    block_t bufMask = block_t(1);

    for (size_t i = 0, n = other.rowCount(); i < n; ++i)
    {
        if (keepRows.find(i) == keepRows.end()) continue;

        const NullableBool b = other.getCell(i);
        if (b != NullableBool::Null)
        {
            bufferNotNA |= bufMask;
            if (b == NullableBool::True) bufferTrue |= bufMask;
        }

        bufMask <<= 1;
        if (bufMask == 0)
        {
            dataTrue.push_back(bufferTrue);
            dataNotNA.push_back(bufferNotNA);
            bufferTrue = bufferNotNA = Zeros<block_t>();
            bufMask = block_t(1);
        }
    }

    if (bufMask != 1)
    {
        dataTrue.push_back(bufferTrue);
        dataNotNA.push_back(bufferNotNA);
    }

    calculateTransientValues();
}

void MineColumn::load(const Table& table, size_t columnIndex, const std::string& valueSet)
{
    // reserve
    const size_t bits = 8*sizeof(block_t);
    size_t dataLength = (table.rows.size() + bits-1) / bits;
    dataTrue.clear();
    dataTrue.reserve(dataLength);
    dataNotNA.clear();
    dataNotNA.reserve(dataLength);

    block_t bufferTrue = Zeros<block_t>(), bufferNotNA = Zeros<block_t>();
    block_t bufMask = block_t(1);

    for (size_t i = 0, n = table.rows.size(); i < n; ++i)
    {
        const char cell = at(at(table.rows, i), columnIndex);
        if (cell != CHAR_NA)
        {
            bufferNotNA |= bufMask;
            if (valueSet.find(cell) != std::string::npos) bufferTrue |= bufMask;
        }

        bufMask <<= 1;
        if (bufMask == 0)
        {
            dataTrue.push_back(bufferTrue);
            dataNotNA.push_back(bufferNotNA);
            bufferTrue = bufferNotNA = Zeros<block_t>();
            bufMask = block_t(1);
        }
    }

    if (bufMask != 1)
    {
        dataTrue.push_back(bufferTrue);
        dataNotNA.push_back(bufferNotNA);
    }

    calculateTransientValues();
}

void MineColumn::calculateTransientValues()
{
    universe = 0;
    for (block_t x : dataNotNA) universe += DefaultPopCount::get(x);
    support = 0;
    for (block_t x : dataTrue) support += DefaultPopCount::get(x);
}

Mine::Mine(const Table& table, const Params& params)
    : Mine()
{
	for (size_t colIndex = 0; colIndex < table.columns.size(); colIndex++)
	{
	    const TableColumn& tableColumn = at(table.columns, colIndex);
	    auto cpIterator = params.columnParams.find(tableColumn.name);
		if (cpIterator == params.columnParams.end())
        {
            cpIterator = params.columnParams.find('#'+::toString(colIndex));
            if (cpIterator == params.columnParams.end())
                continue;
        }
        const ColumnParams& cp = cpIterator->second;

		const std::vector<std::string> valueSets = cp.valueSets->getValueSets
            (std::string(tableColumn.values.begin(), tableColumn.values.end()));

		// All valuesets for this table column will belong to the same group in the mine
		{
            ColumnGroup g;
            g.maxLhsCount = (cp.lhs ? (dynamic_cast<RangeValueSets*>(cp.valueSets.get()) != nullptr ? 2 : 1) : 0);
            f_columnGroups.push_back(std::move(g));
		}

		// create equalities (mine columns)
		for (const std::string& valueSet : valueSets)
		{
			MineColumn mc(tableColumn.name + "=" + valueSet);
			mc.load(table, colIndex, valueSet);
			mc.groupIndex = columnGroups().size() - 1;
			mc.rhsAllowed = cp.rhs;
			f_columns.push_back(std::move(mc));
		}
	}
}

Mine::Mine(const Mine& other, const std::set<size_t>& keepRows)
    : Mine()
{
    f_columns.reserve(other.columns().size());
    for (const MineColumn& otherColumn : other.columns())
        f_columns.emplace_back(otherColumn, keepRows);

    f_columnGroups = other.columnGroups();
}

size_t Mine::rowCount() const
{
    return columns().size() == 0 ? 0 : at(columns(), 0).rowCount();
}

/// universeColumns must be null if and only if ConsiderUniverseColumns is false.
/// The role of ConsiderUniverseColumns is to eliminate extra checks and thus speed up execution.
template <bool ConsiderUniverseColumns>
void _getRuleStats
        (const Mine& mine,
        Rule &rule,
        const double eValueFactor,
        const std::vector<size_t>* const universeColumns)
{
    if ((universeColumns != nullptr) != ConsiderUniverseColumns)
        throw std::invalid_argument("universeColumns must be null iff !ConsiderUniverseColumns");

    rule.stats.clearBasicFields();

    // initialize LHS scanlines
    const size_t nodeCount = rule.lhs.nodes.size();
    std::vector<MineColumn::block_t> const* scanlineTrue[nodeCount];
    std::vector<MineColumn::block_t> const* scanlineNotNA[nodeCount];
    size_t blockCount = 0;
    for (size_t nodeIndex = 0; nodeIndex < nodeCount; ++nodeIndex)
    {
        const Nullable<size_t>& colIndex = at(rule.lhs.nodes, nodeIndex).column;
        if (colIndex.isNull())
        {
            scanlineTrue[nodeIndex] = nullptr;
            scanlineNotNA[nodeIndex] = nullptr;
        }
        else
        {
            const MineColumn& col = at(mine.columns(), (size_t)colIndex);
            scanlineTrue[nodeIndex] = &col.dataTrue;
            scanlineNotNA[nodeIndex] = &col.dataNotNA;
            if (blockCount == 0)
                blockCount = col.dataTrue.size();
        }
    }

    // initialize RHS scanlines
    const MineColumn& rhsColumn = at(mine.columns(), rule.rhsColumnIndex);
    std::vector<MineColumn::block_t> const& rhs_scanlineTrue = rhsColumn.dataTrue;
    std::vector<MineColumn::block_t> const& rhs_scanlineNotNA = rhsColumn.dataNotNA;

    // initialize universe scanlines
    const size_t universeColumnCount = (ConsiderUniverseColumns ? universeColumns->size() : 0);
    std::vector<MineColumn::block_t> const* universe_scanlineNotNA[universeColumnCount];
    if (ConsiderUniverseColumns)
    {
        for (size_t universeColumnIndex = 0; universeColumnIndex < universeColumnCount; ++universeColumnIndex)
        {
            const MineColumn& col = at(mine.columns(), at(*universeColumns, universeColumnIndex));
            universe_scanlineNotNA[universeColumnIndex] = &col.dataNotNA;
        }
    }

    for (size_t i = 0; i < blockCount; ++i)
    {
        // evaluate LHS and iterate
        // get universe
        MineColumn::block_t lhsNotNA = ~static_cast<MineColumn::block_t>(0);
        for (size_t nodeIndex = 0; nodeIndex < nodeCount; ++nodeIndex)
        {
            std::vector<MineColumn::block_t> const* v = scanlineNotNA[nodeIndex];
            if (v == nullptr) continue;

            lhsNotNA &= v->at(i);
        }
        if (ConsiderUniverseColumns)
        {
            for (size_t universeColumnIndex = 0; universeColumnIndex < universeColumnCount; ++universeColumnIndex)
                lhsNotNA &= universe_scanlineNotNA[universeColumnIndex]->at(i);
        }
        // get support
        MineColumn::block_t lhsTrue = lhsNotNA;
        MineColumn::block_t lhsCurrentOrClause = Zeros<MineColumn::block_t>();
        bool emptyClause = true;
        for (size_t nodeIndex = 0; nodeIndex < nodeCount; ++nodeIndex)
        {
            std::vector<MineColumn::block_t> const* v = scanlineTrue[nodeIndex];
            if (v != nullptr)
            {
                lhsCurrentOrClause |= v->at(i);
                emptyClause = false;
            }

            // AND operator?
            if (at(rule.lhs.nodes, nodeIndex).op == ConjunctiveNormalForm::Operator::And ||
                nodeIndex == nodeCount-1)
            {
                // do not take empty OR clauses into account
                if (!emptyClause)
                    lhsTrue &= lhsCurrentOrClause;

                lhsCurrentOrClause = Zeros<MineColumn::block_t>();
                emptyClause = true;
            }
        }

        // evaluate RHS and iterate
        // get universe
        const MineColumn::block_t rhsNotNA = at(rhs_scanlineNotNA, i);
        // get support
        const MineColumn::block_t rhsTrue = at(rhs_scanlineTrue, i);

        // update rule stats
        rule.stats.universe += DefaultPopCount::get(lhsNotNA & rhsNotNA);
        rule.stats.supportLhs += DefaultPopCount::get(lhsTrue & rhsNotNA);
        rule.stats.supportRhs += DefaultPopCount::get(lhsNotNA & rhsTrue);
        rule.stats.support += DefaultPopCount::get(lhsTrue & rhsTrue);
    }

    rule.stats.calculateTransientValues(eValueFactor);
}

void Mine::getRuleStats(Rule& rule, const double eValueFactor) const
{
    _getRuleStats<false>(*this, rule, eValueFactor, nullptr);
}

void Mine::getRuleStats(Rule& rule, const double eValueFactor,
        const std::vector<size_t>* const universeColumns) const
{
    _getRuleStats<true>(*this, rule, eValueFactor, universeColumns);
}
