/*
SCARF Association Rule Mining Program
Copyright (c) 2012-13 Balazs Szalkai

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#pragma once

#include <cstddef>
#include <stdexcept>
#include <utility>

/// T should be a type with simple initialization, like a primitive type, string or vector.
template <class T>
class Nullable
{
    private: bool f_isNull;
    private: T value;

    public: Nullable()
        : f_isNull(true), value()
    {
    }

    public: Nullable(const T& _value)
        : f_isNull(false), value(_value)
    {
    }

    public: Nullable(const Nullable& other) = default;

    public: Nullable(T&& _value)
        : f_isNull(false), value(std::move(_value))
    {
    }

    public: Nullable(Nullable&& other) = default;

    public: bool isNull() const
    {
        return f_isNull;
    }

    public: void setNull()
    {
        f_isNull = true;
        value = T(); // free memory
    }

    public: Nullable& operator=(const T& _value)
    {
        f_isNull = false;
        value = _value;
        return *this;
    }

    public: Nullable& operator=(const Nullable& other) = default;

    public: Nullable& operator=(T&& _value)
    {
        f_isNull = false;
        value = std::move(_value);
        return *this;
    }

    public: Nullable& operator=(Nullable&& other) = default;

    public: bool operator==(const T& other) const
    {
        if (f_isNull) return false;
        return value == other;
    }

    public: bool operator!=(const T& other) const
    {
        return !(*this == other);
    }

    public: bool operator==(const Nullable& other) const
    {
        if (f_isNull != other.f_isNull) return false;
        if (f_isNull) return true;
        return value == other.value;
    }

    public: bool operator!=(const Nullable& other) const
    {
        return !(*this == other);
    }

    public: bool operator<(const Nullable& other) const
    {
        if (f_isNull) return !other.f_isNull;
        if (other.f_isNull) return false;
        return value < other.value;
    }

    public: explicit operator T& ()
    {
        if (f_isNull) throw std::invalid_argument("Nullable is null");
        return value;
    }

    public: explicit operator const T& () const
    {
        if (f_isNull) throw std::invalid_argument("Nullable is null");
        return value;
    }
};

/// Specialized, faster version for bool
enum class NullableBool
{
    Null,
    False,
    True
};
