/*
SCARF Association Rule Mining Program
Copyright (c) 2012-13 Balazs Szalkai

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#pragma once

#include "concurrent_stream.hpp"
#include <map>
#include <memory>
#include "rulestats.hpp"
#include <string>
#include <typeinfo>
#include <vector>

class IValueSets
{
    public: virtual std::vector<std::string> getValueSets(const std::string& values) = 0;
    public: virtual ~IValueSets() {}
};

/// All possible subsets
class AllValueSets : virtual public IValueSets
{
    public: virtual std::vector<std::string> getValueSets(const std::string&);
};

/// Only <=c or >=c ranges
/// An ordering parameter can be given, which is a string
/// containing the possible values listed from smallest to largest.
/// Values not in "ordering" are considered smaller than those that are,
/// and are ordered by ASCII code.
class RangeValueSets : virtual public IValueSets
{
    private: std::string ordering;

    public: RangeValueSets() {}
    public: RangeValueSets(const std::string& _ordering)
        : ordering(_ordering) {}
    public: int compareValues(char, char);
    public: virtual std::vector<std::string> getValueSets(const std::string&);
};

/// Only one-element subsets
class SingleValueSets : virtual public IValueSets
{
    public: virtual std::vector<std::string> getValueSets(const std::string&);
};

/// Specifies how to handle a data column when mining.
class ColumnParams
{
	public: std::string name;
	public: std::unique_ptr<IValueSets> valueSets;
	// Can the column occur on LHS/RHS?
	public: bool lhs, rhs;

    public: ColumnParams()
        : name(), valueSets(new SingleValueSets), lhs(false), rhs(false)
    {
    }

    /// Parses a definition.
	public: explicit ColumnParams(const std::string&, const ConcurrentStream& logger);
};

template <class Out>
Out& operator<<(Out& out, const ColumnParams& params)
{
    out<<"name = "<<params.name<<", ";
    out<<"valueSets = "<<typeid(*params.valueSets).name()<<", ";
    out<<"lhs = "<<params.lhs<<", ";
    out<<"rhs = "<<params.rhs<<std::endl;
    return out;
}

// Parameters for the association mining program
class Params
{
	/// Table file path (.csv file, absolute path)
	public: std::string tableFile;
	/// Output file path (absolute path), or "" if output should be written on stdout
	public: std::string outputFile;
	/// XML output file path, "" means no XML output.
	public: std::string xmlOutputFile;
	/// Number of data mining threads (0: fall back to default)
	public: size_t nThreads;
	/// Pattern string for mining rules
	/// Should contain the sizes of OR clauses,
	/// e.g. "1 2". This will result in something like 'age=B & (race=W or lb_bpsys=h)'
	public: std::string patternStr;
	/// Number of rules that need to be mined
	public: size_t maxRuleCount;
	/// Minimum statistics for rule mining
	public: RuleStats minStats;
	/// Allowed goodness decrease when simplifying rules
	public: double allowedGoodnessDecrease;
	/// Maps database column names to column parameters
	public: std::map<std::string, ColumnParams> columnParams;

    public: explicit Params(const std::string &fn, const ConcurrentStream& logger);
};

template <class Out>
Out& operator<<(Out& out, const Params& params)
{
    out<<"tableFile = "<<params.tableFile<<std::endl;
    out<<"outputFile = "<<params.outputFile<<std::endl;
    out<<"xmlOutputFile = "<<params.xmlOutputFile<<std::endl;
    out<<"nThreads = "<<params.nThreads<<std::endl;
    out<<"patternStr = "<<params.patternStr<<std::endl;
    out<<"maxRuleCount = "<<params.maxRuleCount<<std::endl;
    out<<"minStats:"<<std::endl;
    out<<params.minStats.toString();
    out<<"allowedGoodnessDecrease = "<<params.allowedGoodnessDecrease<<std::endl;
    out<<"columnParams:"<<std::endl;
    for (auto it = params.columnParams.begin(); it != params.columnParams.end(); ++it)
        out<<it->second;
    return out;
}
