/*
SCARF Association Rule Mining Program
Copyright (c) 2012-13 Balazs Szalkai

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "rule.hpp"

#include <algorithm>
#include <cstdint>
#include <fstream>
#include <iostream>
#include "mine.hpp"
#include <utility>
#include "utils.hpp"

std::int64_t Rule::getHash() const
{
    return (lhs.getHash()*HASH_PRIME) ^ rhsColumnIndex;
}

std::string Rule::toString(const Mine &mine) const
{
    const std::string& rhsColumnName = at(mine.columns(), rhsColumnIndex).name;
    return lhs.toString(mine)
        +" ---> "
        +rhsColumnName
        +"\n"
        +stats.toString();
}

void Rule::toXml(XmlWriter& xml, const Mine &mine) const
{
    const std::string& rhsColumnName = at(mine.columns(), rhsColumnIndex).name;
    xml.start("rule");
    xml.start("expression");
    xml.start("implies");
    lhs.toXml(xml, mine);
    xml.start("variable");
    xml.text(rhsColumnName);
    xml.end();
    xml.end();
    xml.end();
    stats.toXml(xml);
    xml.end();
}

/// WARNING: This class has reference fields.
template <class T>
class Leaf
{
    public: T& location;
    public: const T oldValue;

    public: explicit Leaf(T& value)
        : location(value), oldValue(value) {}

    public: void reset() const
    {
        location = oldValue;
    }
};

bool Rule::canSimplify(const Mine &mine, const RuleStats &minStats,
                         const double eValueFactor, const double allowedGoodnessDecrease) const
{
    Rule other = *this;

    // Extract the leaves of the left-hand side
    std::vector< Leaf<Nullable<size_t>> > leaves;
    for (auto& node : other.lhs.nodes)
        if (!node.column.isNull()) leaves.push_back(Leaf<Nullable<size_t>>(node.column));
    //if (leaves.size() <= 1) return false;

    std::vector<size_t> deletedColumns;
    uint64_t iMax = (static_cast<uint64_t>(1) << std::min(static_cast<size_t>(63), leaves.size()));
    // i = 0 would mean preserving the orig. rule completely
    for (uint64_t i = 1; i < iMax; i++)
    {
        // decode i
        uint64_t code = i;
        deletedColumns.clear();
        for (size_t j = 0; code != 0; ++j)
        {
            // delete the leaf?
            if ((code & 1u) != 0)
            {
                const Leaf<Nullable<size_t>>& leaf = at(leaves, j);
                if (!leaf.oldValue.isNull())
                {
                    leaf.location.setNull();
                    deletedColumns.push_back((size_t)leaf.oldValue);
                }
            }
            // shift
            code >>= 1u;
        }

        const double q = 1.0 - (1.0-allowedGoodnessDecrease)*deletedColumns.size();

        // calculate new statistics
        // keep universe of original rule
        mine.getRuleStats(other, eValueFactor, &deletedColumns);
        if (other.stats.goodness() >= stats.goodness() * q && other.stats.atLeastAsGoodAs(minStats))
            return true;
        // take universe of the new rule
        mine.getRuleStats(other, eValueFactor);
        if (other.stats.goodness() >= stats.goodness() * q && other.stats.atLeastAsGoodAs(minStats))
            return true;

        // restore previous contents of leaves
        for (const Leaf<Nullable<size_t>>& leaf : leaves)
            leaf.reset();
    }

    return false;
}
