/*
SCARF Association Rule Mining Program
Copyright (c) 2012-13 Balazs Szalkai

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#pragma once

#include <cassert>
#include <cstdint>
#include <iostream>
#include "range.hpp"
#include <sstream>
#include <stdexcept>
#include <string>
#include <vector>

// If NDEBUG is defined, then this is fast (equivalent to v[index])
template <class V>
typename V::reference at(V& v, typename V::size_type index)
{
    assert(index < v.size());
    return v[index];
}

// If NDEBUG is defined, then this is fast (equivalent to v[index])
template <class V>
typename V::const_reference at(const V& v, typename V::size_type index)
{
    assert(index < v.size());
    return v[index];
}

const int64_t HASH_PRIME = 1000000007;
const int64_t HASH_NULL = HASH_PRIME;

const char CHAR_NA = 0; // should not be defined here but in some other header

template <class T> int compare (const T a, const T b)
{
    return a < b ? -1 : b < a ? 1 : 0;
}

template <class T> T sqr(const T x) {return x*x;}

constexpr std::int64_t getHash(const char *s)
{
    return !*s ? 11 : *s ^ 7*getHash(s+1);
}

std::int64_t getHash(const std::string &s);

bool isWhitespace(char c);

/// Throws on unsuccessful parse.
constexpr std::int64_t parseInt(const char* const s, const std::int64_t currentValue = 0)
{
    return !*s ? currentValue :
        *s < '0' || *s > '9' ? throw std::invalid_argument("Cannot parse integer.") :
        parseInt(s+1, currentValue*10 + (*s-'0'));
}

std::int64_t parseInt(const std::string &);

template <class T>
std::string toString(const T arg)
{
    std::ostringstream s;
    s<<arg;
    return s.str();
}

enum class EmptyEntryAction
{
    Keep,
    Remove
};

template <class Char>
class const_string
{
    private: const Char* const data;
    private: const size_t length;

    private: static constexpr size_t getLength(const Char* _data)
    {
        return *_data ? 1 + getLength(_data + 1) : 0;
    }

    public: constexpr const_string(const Char* _data)
        : data(_data), length(getLength(_data)) {}

    public: operator const Char* () const { return data; }

    public: const Char* begin() const { return data; }
    public: const Char* end() const { return data + length; }
    public: size_t size() const { return length; }
};

/// Splits a string
template <class Predicate>
    std::vector<std::string> split(const std::string &s, const Predicate is_separator, const EmptyEntryAction a)
{
    std::vector<std::string> result;
    std::string::const_iterator start = s.begin();
    for (std::string::const_iterator it = s.begin(); it != s.end(); ++it)
    {
        if (is_separator(*it))
        {
            if (a == EmptyEntryAction::Keep || start != it)
                result.push_back(std::string(start, it));
            start = it+1;
        }
    }
    if (a == EmptyEntryAction::Keep || start != s.end())
        result.push_back(std::string(start, s.end()));
    return result;
}

std::vector<std::string> split(const std::string &s, const char separator, const EmptyEntryAction a);
std::vector<std::string> csvSplit(const std::string &s, const char separator);

template <class It1, class It2>
bool startsWith(const It1 haystackBegin, const It1 haystackEnd, const It2 needleBegin, const It2 needleEnd)
{
    if (haystackEnd-haystackBegin < needleEnd-needleBegin) return false;
    for (auto ith = haystackBegin, itn = needleBegin; itn != needleEnd; ++ith, ++itn)
        if (*ith != *itn) return false;
    return true;
}

template <class Container1, class Container2>
bool startsWith(const Container1& haystack, const Container2& needle)
{
    return startsWith(haystack.begin(), haystack.end(), needle.begin(), needle.end());
}

/// Trims leading and trailing whitespaces
std::pair<std::string::const_iterator,std::string::const_iterator> stripToRange(const std::string &s);
std::string strip(const std::string &s);
